import subprocess
from fnmatch import fnmatch
from pathlib import Path
from typing import Callable

import yaml

from myteam.templates import get_template


def get_myteam_root(cur_dir: Path):
    d = cur_dir
    while d.parent != d:  # i.e. not at root
        if d.name == '.myteam':
            return d
        d = d.parent
    return cur_dir


def _print_block(text: str) -> None:
    print(text.rstrip('\n') + '\n')


def _strip_yaml_frontmatter(text: str) -> str:
    lines = text.splitlines()
    if not lines or lines[0].strip() != "---":
        return text

    for i in range(1, len(lines)):
        if lines[i].strip() == "---":
            body = "\n".join(lines[i + 1:])
            if text.endswith("\n"):
                body += "\n"
            return body
    return text


def print_instructions(base: Path):
    for file in ['role.md', 'ROLE.md', 'skill.md', 'SKILL.md']:
        instructions_file = base / file
        if instructions_file.exists():
            _print_block(_strip_yaml_frontmatter(instructions_file.read_text(encoding='utf-8')))
            return


def _get_definition_file(folder: Path, definition_stem: str) -> Path | None:
    for candidate in (f"{definition_stem}.md", f"{definition_stem.upper()}.md"):
        definition_file = folder / candidate
        if definition_file.exists():
            return definition_file
    return None


def is_role_dir(folder: Path) -> bool:
    return folder.is_dir() and _get_definition_file(folder, "role") is not None


def is_skill_dir(folder: Path) -> bool:
    return folder.is_dir() and _get_definition_file(folder, "skill") is not None


def _parse_yaml_frontmatter(file: Path) -> dict[str, str]:
    if not file.exists():
        return {}

    lines = file.read_text(encoding="utf-8").splitlines()
    if not lines or lines[0].strip() != "---":
        return {}

    end = None
    for i in range(1, len(lines)):
        if lines[i].strip() == "---":
            end = i
            break
    if end is None:
        return {}

    frontmatter = "\n".join(lines[1:end])
    try:
        loaded = yaml.safe_load(frontmatter)
    except yaml.YAMLError:
        return {}

    if not isinstance(loaded, dict):
        return {}

    data: dict[str, str] = {}
    for key, value in loaded.items():
        if value is None:
            continue
        data[str(key).lower()] = str(value)

    return data


def _format_frontmatter_info(frontmatter: dict[str, str]) -> str:
    name = frontmatter.get("name", "")
    description = frontmatter.get("description", "")
    if name and description:
        return f"{name}: {description}"
    if name:
        return name
    if description:
        return description
    return ""


def _get_folder_info(folder: Path, definition_stem: str) -> str:
    definition_file = _get_definition_file(folder, definition_stem)
    if definition_file is not None:
        frontmatter_info = _format_frontmatter_info(_parse_yaml_frontmatter(definition_file))
        if frontmatter_info:
            return frontmatter_info

    info = folder / "info.md"
    if info.exists():
        return info.read_text(encoding="utf-8").rstrip('\n')
    return ""


def _is_py_file(file: Path) -> bool:
    return file.is_file() and file.suffix == '.py'


def _print_info(
        header: str,
        folder: Path, base_dir: Path, ignore: list[str],
        is_relevant: Callable[[Path], bool],
        get_info: Callable[[Path], str],
):
    relevant = list(sorted(
        p
        for p in folder.iterdir()
        if is_relevant(p) and p.name not in ignore
    ))
    if not relevant:
        return

    print()
    print(f' {header} '.center(30, '*'))
    for cur_dir in relevant:
        name = cur_dir.relative_to(base_dir).as_posix()
        print(f" {name} ".center(30, '-'))
        if info := get_info(cur_dir):
            print(info)
    print()


def _matches_tree_glob(path: Path, root: Path, glob: str) -> bool:
    return path.relative_to(root).match(glob)


def _is_excluded_tree_path(path: Path, root: Path, exclude: tuple[str, ...]) -> bool:
    relative_path = path.relative_to(root).as_posix()
    return any(
        fnmatch(path.name, pattern) or fnmatch(relative_path, pattern)
        for pattern in exclude
    )


def _get_git_ignored_paths(root: Path) -> set[str]:
    try:
        repo_root_text = subprocess.check_output(
            ["git", "-C", str(root), "rev-parse", "--show-toplevel"],
            stderr=subprocess.DEVNULL,
            text=True,
        ).strip()
    except (OSError, subprocess.CalledProcessError):
        return set()

    repo_root = Path(repo_root_text)
    try:
        relative_root = root.relative_to(repo_root)
    except ValueError:
        return set()

    pathspec = relative_root.as_posix()
    if not pathspec:
        pathspec = "."

    try:
        output = subprocess.check_output(
            [
                "git",
                "-C",
                str(repo_root),
                "ls-files",
                "--ignored",
                "--exclude-standard",
                "--others",
                "--directory",
                pathspec,
            ],
            stderr=subprocess.DEVNULL,
            text=True,
        )
    except (OSError, subprocess.CalledProcessError):
        return set()

    ignored: set[str] = set()
    prefix = "" if pathspec == "." else f"{pathspec.rstrip('/')}/"
    for line in output.splitlines():
        normalized = line.rstrip("/")
        if prefix and normalized.startswith(prefix):
            normalized = normalized.removeprefix(prefix)
        ignored.add(normalized)
    return ignored


def _is_git_ignored_tree_path(path: Path, root: Path, ignored_paths: set[str]) -> bool:
    relative_path = path.relative_to(root).as_posix()
    for ignored in ignored_paths:
        if relative_path == ignored or relative_path.startswith(f"{ignored}/"):
            return True
    return False


def _collect_tree_entries(
        root: Path,
        folder: Path,
        glob: str,
        max_levels: int | None,
        level: int,
        exclude: tuple[str, ...],
        ignored_paths: set[str],
) -> list[tuple[Path, list[tuple[Path, list]]]]:
    entries: list[tuple[Path, list[tuple[Path, list]]]] = []
    children = sorted(folder.iterdir(), key=lambda path: (not path.is_dir(), path.name.lower(), path.name))

    for child in children:
        if _is_excluded_tree_path(child, root, exclude):
            continue
        if ignored_paths and _is_git_ignored_tree_path(child, root, ignored_paths):
            continue

        if child.is_dir():
            if max_levels is not None and level > max_levels:
                continue

            descendants: list[tuple[Path, list[tuple[Path, list]]]] = []
            if max_levels is None or level < max_levels:
                descendants = _collect_tree_entries(root, child, glob, max_levels, level + 1, exclude, ignored_paths)

            if descendants or glob == "*":
                entries.append((child, descendants))
            continue

        if max_levels is not None and level > max_levels:
            continue
        if _matches_tree_glob(child, root, glob):
            entries.append((child, []))

    return entries


def _print_tree_entries(entries: list[tuple[Path, list[tuple[Path, list]]]], prefix: str) -> None:
    for index, (path, children) in enumerate(entries):
        is_last = index == len(entries) - 1
        branch = "\\-- " if is_last else "|-- "
        extension = "    " if is_last else "|   "
        print(f"{prefix}{branch}{path.name}")
        if children:
            _print_tree_entries(children, prefix + extension)


def print_directory_tree(
        root: Path,
        glob: str = "*",
        max_levels: int | None = None,
        exclude: tuple[str, ...] = (".*", "_*"),
        use_gitignore: bool = True,
        relative_to: Path | None = None,
) -> None:
    root = root.resolve()
    header = root.name or root.as_posix()
    if relative_to is not None:
        header = root.relative_to(relative_to.resolve()).as_posix()

    print(header)
    ignored_paths = _get_git_ignored_paths(root) if use_gitignore else set()
    _print_tree_entries(
        _collect_tree_entries(root, root, glob, max_levels, 1, exclude, ignored_paths),
        "",
    )
    print()


def list_roles(folder: Path, base_dir: Path, ignore: list[str]):
    _print_info(
        'Team Members',
        folder,
        base_dir,
        ignore,
        is_role_dir,
        lambda role_dir: _get_folder_info(role_dir, "role"),
    )


def list_skills(folder: Path, base_dir: Path, ignore: list[str]):
    _print_info(
        'Skills',
        folder,
        base_dir,
        ignore,
        is_skill_dir,
        lambda skill_dir: _get_folder_info(skill_dir, "skill"),
    )


def list_tools(folder: Path, base_dir: Path, ignore: list[str]):
    _print_info('Tools', folder, base_dir, ignore + ['load.py'], _is_py_file, lambda f: '')
    # TODO - get usage from py file


def explain_skills():
    _print_block(get_template('explain_skills.md'))


def explain_roles():
    _print_block(get_template('explain_roles.md'))

def explain_tools():
    _print_block(get_template('explain_tools.md'))



"""Path helpers and constants for the myteam CLI."""
from __future__ import annotations

from pathlib import Path

APP_NAME = "myteam"
AGENTS_DIRNAME = ".myteam"
ENCODING = "utf-8"


def base_dir() -> Path:
    """Return the directory from which the CLI was invoked."""
    return Path.cwd()


def agents_root(base: Path) -> Path:
    return base / AGENTS_DIRNAME


def role_dir(base: Path, role: str) -> Path:
    return agents_root(base) / role



"""Myteam package for managing agent role directories."""
from __future__ import annotations

from importlib.metadata import PackageNotFoundError, version
from pathlib import Path
import tomllib

__all__ = ["__version__"]


def _project_version() -> str:
    pyproject = Path(__file__).resolve().parents[2] / "pyproject.toml"
    if pyproject.exists():
        with pyproject.open("rb") as handle:
            project = tomllib.load(handle)["project"]
        return str(project["version"])

    try:
        return version("myteam")
    except PackageNotFoundError:
        raise RuntimeError("Unable to determine myteam version") from None


__version__ = _project_version()



"""Command implementations for the myteam CLI."""
from __future__ import annotations

import shutil
import subprocess
import sys
from pathlib import Path

from myteam.utils import is_role_dir, is_skill_dir

from . import __version__
from .paths import APP_NAME, ENCODING, agents_root, base_dir, role_dir
from .rosters import download_roster, list_available_rosters
from .templates import get_template


def ensure_dir(path: Path) -> None:
    path.mkdir(parents=True, exist_ok=True)


def write_py_script(path: Path, contents: str) -> None:
    path.write_text(contents, encoding=ENCODING)


def new_dir(
    base: Path,
    dir_type: str,
    name_parts: list[str],
    instruction_text: str,
    load_text: str,
) -> None:
    name_dir = base.joinpath(*name_parts)
    name = "/".join(name_parts)
    if name_dir.exists():
        print(f"{dir_type.title()} '{name}' already exists at {name_dir}", file=sys.stderr)
        raise SystemExit(1)

    ensure_dir(name_dir)
    (name_dir / f"{dir_type}.md").write_text(instruction_text, encoding=ENCODING)
    write_py_script(name_dir / "load.py", load_text)


def init() -> None:
    """Initialize the myteam directory with default role."""
    new_dir(
        base_dir(),
        "role",
        [".myteam"],
        "",
        get_template("role_load_template.py"),
    )

    agents_md = base_dir() / "AGENTS.md"
    if not agents_md.exists():
        agents_md.write_text(get_template("agents_md_template.md"), encoding=ENCODING)


def new_role(role: str) -> None:
    """Create a new role directory with placeholder files."""
    new_dir(
        agents_root(base_dir()),
        "role",
        role.split("/"),
        get_template("role_definition_template.md"),
        get_template("role_load_template.py"),
    )


def new_skill(skill: str) -> None:
    new_dir(
        agents_root(base_dir()),
        "skill",
        skill.split("/"),
        get_template("skill_definition_template.md"),
        get_template("skill_load_template.py"),
    )


def remove(name: str) -> None:
    """Delete the directory for a role or skill if it exists."""
    target_dir = role_dir(base_dir(), name)  # TODO fix for skills
    if not target_dir.exists():
        print(f"'{name}' not found at {target_dir}", file=sys.stderr)
        raise SystemExit(1)

    if not target_dir.is_dir():
        print(f"Path for '{name}' is not a directory: {target_dir}", file=sys.stderr)
        raise SystemExit(1)

    try:
        shutil.rmtree(target_dir)
    except OSError as exc:
        print(f"Failed to remove '{name}': {exc}", file=sys.stderr)
        raise SystemExit(1)


def get_name(dir_type: str, name_dir: Path, name: str | None) -> None:
    if not name_dir.exists():
        print(
            f"{dir_type.title()} '{name}' not found. Run 'myteam new {dir_type} {name}' to create it.",
            file=sys.stderr,
        )
        raise SystemExit(1)

    load_py = name_dir / "load.py"
    if not load_py.exists():
        print(f"No load.py found for {dir_type} '{name}'.", file=sys.stderr)
        raise SystemExit(1)

    try:
        result = subprocess.run([sys.executable, str(load_py)], cwd=name_dir, check=False)
        raise SystemExit(result.returncode)
    except OSError as exc:
        print(f"Failed to execute load.py for {dir_type} '{name}': {exc}", file=sys.stderr)
        raise SystemExit(1)


def get_role(role: str | None = None) -> None:
    """Print the instructions for the given role if available."""
    folder = agents_root(base_dir())
    if role is not None:
        folder = folder.joinpath(*role.split("/"))

    if not is_role_dir(folder):
        print(f"Not a role: {role}", file=sys.stderr)
        raise SystemExit(1)
    get_name("role", folder, role)


def get_skill(skill: str) -> None:
    """Print the instructions for the given skill if available."""
    folder = agents_root(base_dir()).joinpath(*skill.split("/"))
    if not is_skill_dir(folder):
        print(f"Not a skill: {skill}", file=sys.stderr)
        raise SystemExit(1)
    get_name("skill", folder, skill)


def version() -> str:
    return f"{APP_NAME} {__version__}"


__all__ = [
    "download_roster",
    "get_role",
    "get_skill",
    "init",
    "list_available_rosters",
    "new_role",
    "new_skill",
    "remove",
    "version",
]



"""Embedded template assets for myteam."""
from pathlib import Path

_template_dir = Path(__file__).parent


def get_template(name: str) -> str:
    template = _template_dir / name
    return template.read_text()



#!/usr/bin/env python3
from __future__ import annotations

from pathlib import Path

from myteam.utils import print_instructions, get_myteam_root, list_roles, list_skills, list_tools


def main() -> int:
    base = Path(__file__).resolve().parent  # .myteam/<role>
    print_instructions(base)
    myteam = get_myteam_root(base)
    list_roles(base, myteam, [])
    list_skills(base, myteam, [])
    list_tools(base, myteam, [])

    return 0


if __name__ == "__main__":
    raise SystemExit(main())



"""Command-line interface wiring for the myteam package."""
from __future__ import annotations

import fire

from .commands import (
    download_roster,
    get_role,
    get_skill,
    init,
    list_available_rosters,
    new_role,
    new_skill,
    remove,
    version,
)


def main(argv: list[str] | None = None):
    commands = {
        "init": init,
        "new": {
            "role": new_role,
            "skill": new_skill
        },
        "remove": remove,
        "get": {
            "role": get_role,
            "skill": get_skill,
        },
        "download": download_roster,
        "list": list_available_rosters,
        "--version": version,
    }

    fire.Fire(commands)


if __name__ == "__main__":
    raise SystemExit(main())



#!/usr/bin/env python3
from __future__ import annotations

from pathlib import Path

from myteam.utils import print_instructions, get_myteam_root, explain_skills, explain_roles, explain_tools, list_skills, \
    list_roles, list_tools, print_directory_tree


def main() -> int:
    base = Path(__file__).resolve().parent  # .myteam/<role>
    myteam = get_myteam_root(base)

    print_instructions(base)
    print_directory_tree(myteam.parent)

    explain_roles()
    list_roles(base, myteam, [])

    explain_skills()
    list_skills(base, myteam, [])

    explain_tools()
    list_tools(base, myteam, [])

    return 0


if __name__ == "__main__":
    raise SystemExit(main())



"""Roster listing and download logic."""
from __future__ import annotations

import json
import sys
import urllib.request
from pathlib import Path

APP_NAME = "myteam"
AGENTS_DIRNAME = ".myteam"
DEFAULT_REPO = "beanlab/rosters"


def _agents_root(base: Path) -> Path:
    return base / AGENTS_DIRNAME


def _repo_urls(repo: str) -> tuple[str, str]:
    repo_path = repo.strip().strip("/")
    if repo_path.count("/") != 1:
        print(f"Invalid repo '{repo}'. Expected format: <owner>/<repo>.", file=sys.stderr)
        exit(1)
    api_base = f"https://api.github.com/repos/{repo_path}/git/trees"
    raw_base = f"https://raw.githubusercontent.com/{repo_path}/refs/heads/main"
    return api_base, raw_base


def _fetch_json(url: str) -> dict:
    try:
        request = urllib.request.Request(url, headers={"User-Agent": APP_NAME})
        with urllib.request.urlopen(request) as response:
            return json.load(response)
    except Exception as exc:
        print(f"Failed to fetch JSON from {url}: {exc}", file=sys.stderr)
        exit(1)


def _download_file(url: str, output_path: Path):
    try:
        request = urllib.request.Request(url, headers={"User-Agent": APP_NAME})
        with urllib.request.urlopen(request) as response:
            data = response.read()
        output_path.parent.mkdir(parents=True, exist_ok=True)
        output_path.write_bytes(data)
    except Exception as exc:
        print(f"Failed to download file from {url}: {exc}", file=sys.stderr)
        exit(1)


def _fetch_available_rosters(roster_repository_url: str):
    root_tree = _fetch_json(roster_repository_url + "/main?recursive=1")
    return root_tree.get("tree", [])


def _fetch_roster_tree(roster: str, roster_repository_url: str):
    roster_trees = _fetch_available_rosters(roster_repository_url)
    roster_tree = next(
        (entry for entry in roster_trees if entry.get("path") == roster),
        None,
    )
    if roster_tree is None:
        _handle_missing_roster(roster, roster_trees)

    return roster_tree


def _handle_missing_roster(roster: str, roster_trees: list):
    roster_names = [entry.get("path") for entry in roster_trees if entry.get("type") == "tree"]
    roster_names.sort()
    available = ", ".join(roster_names) if roster_names else "none"
    print(f"Roster '{roster}' not found. Available rosters: {available}", file=sys.stderr)
    exit(1)


def _fetch_tree_files(roster_tree, roster_repository_url: str):
    subtree_url = f"{roster_repository_url}/{roster_tree['sha']}?recursive=1"
    subtree = _fetch_json(subtree_url)
    file_entries = [entry for entry in subtree.get("tree", []) if entry.get("type") == "blob"]
    if not file_entries:
        print(f"No files found in roster '{roster_tree.get('path')}'.", file=sys.stderr)
        exit(1)

    return file_entries


def _download_blob(blob_object: dict, destination: Path, roster_raw_base_url: str):
    file_name = blob_object.get('path').split("/")[-1]
    print(f"\rDownloading {file_name}")
    _download_file(f"{roster_raw_base_url}/{blob_object.get('path')}", destination / file_name)


def _download_tree_files(file_entries, roster_dir_name: str, destination: Path, roster_raw_base_url: str):
    total = len(file_entries)
    for idx, entry in enumerate(file_entries, start=1):
        rel_path = entry.get("path")
        if not rel_path:
            continue
        raw_url = f"{roster_raw_base_url}/{roster_dir_name}/{rel_path}"
        print(f"\rDownloading {roster_dir_name} {idx}/{total}", end="", file=sys.stderr)
        _download_file(raw_url, destination / rel_path)
    print("", file=sys.stderr)


def download_roster(
    roster_dir_name: str,
    destination: Path | str | None = None,
    repo: str = DEFAULT_REPO,
):
    base = Path.cwd()
    if destination is None:
        destination = _agents_root(base)
    else:
        destination = Path(destination)

    roster_repository_url, roster_raw_base_url = _repo_urls(repo)
    roster_tree = _fetch_roster_tree(roster_dir_name, roster_repository_url)
    if roster_tree.get('type') == 'blob':
        _download_blob(roster_tree, destination, roster_raw_base_url)
    else:
        tree_files = _fetch_tree_files(roster_tree, roster_repository_url)
        _download_tree_files(tree_files, roster_dir_name, destination, roster_raw_base_url)


def list_available_rosters(repo: str = DEFAULT_REPO):
    roster_repository_url, _ = _repo_urls(repo)
    available_rosters = _fetch_available_rosters(roster_repository_url)
    roster_names = [roster.get("path") for roster in available_rosters]
    for name in roster_names:
        print(name)
